#! /home/alergn/virtualenvs/torchgpu/bin/python3.9
import click
import sys
import yaml

@click.command()
@click.option('--config_file', type=str)
def training(config_file):
    
    # Open config file
    with open(config_file) as file:
        config = yaml.load(file, Loader=yaml.FullLoader)

    # Import script (this changes based on which config file arg is passed by the user)
    module_name = config['training_script']
    __import__(module_name)

    # Make imported module variables available
    imported_module = sys.modules[module_name]

    # Create new training object
    train_obj = imported_module.Training(config_file)

    # *** EXPERIMENT CHOICE ***
    # TRAINING NOMAD
    if config['experiment_name'] == 'Training':
        train_obj.training_loop()
    
    # PERFORMANCE EVALUATION
    # NMR AUDIO QUALITY
    elif config['experiment_name'] == 'quality_nmr':
        train_obj.eval_audio_quality(config['nomad_model_path'])
    
    # NMR RANKING VALIDATION SET CONDITIONS 
    elif config['experiment_name'] == 'valid_rank':
        train_obj.eval_degr_level(config['nomad_model_path'])
    
    # NMR RANKING DEGRADATION INTENSITY
    elif config['experiment_name'] == 'intensity':
        train_obj.eval_degradation_intensity(config['nomad_model_path'])

    # FULL REFERENCE AUDIO QUALITY
    elif config['experiment_name'] == 'quality_fr':
        train_obj.eval_full_reference(config['nomad_model_path'])

    # PLOT PERFORMANCE (to check)
    elif config['experiment_name'] == 'Plot':
        train_obj.plot_performance()

if __name__ == '__main__':
    training()  


from nomad import nomad

def test_nomad_score():
    nmr_path = 'data/nmr-data'
    test_path = 'data/test-data'

    avg_score, scores = nomad.predict(nmr_path, test_path)
    
    print(avg_score)
    print('\n')
    print(scores)


import math
import subprocess
import numpy as np
import soundfile as sf
import os
import torchaudio

def mp3(in_filepath, out_filepath, bitrate='320k', sr=16000):
    # Compress mp3
    mp3_outpath = os.path.splitext(out_filepath)[0] + '.mp3'
    cmd = f'ffmpeg -y -i {in_filepath} -ar {sr} -b:a {bitrate} {mp3_outpath}'.split(' ')
    subprocess.call(cmd)

    # Convert to wav 
    wav_convert = f'ffmpeg -y -i {mp3_outpath} -ar {sr} {out_filepath}'.split(' ')
    subprocess.call(wav_convert)
    os.remove(mp3_outpath)
    
def opus(in_filepath, out_filepath, bitrate='320k', sr=16000):
    # Compress opus
    opus_outpath = os.path.splitext(out_filepath)[0] + '.opus'
    cmd = f'ffmpeg -y -i {in_filepath} -c:a libopus -b:a {bitrate} -vbr on {opus_outpath}'.split(' ')
    subprocess.call(cmd)

    # Convert to wav
    wav_convert = f'ffmpeg -y -i {opus_outpath} -ar {sr} {out_filepath}'.split(' ')
    subprocess.call(wav_convert)
    os.remove(opus_outpath)
    
def noise(clean_path, noise_path, out_filepath, snr_db=0, sr=16000):
    x, sr = sf.read(clean_path)
    s, sr = sf.read(noise_path)

    # First let's match the length of the two signals
    x_len = x.shape[0]
    s_len = s.shape[0]

    if x_len > s_len:
        # Calculate how many times we need to repeat the signal and round up to the nearest integer
        reps = math.ceil(x_len/s_len)

        # Use the function np.tile to repeat an array
        s = np.tile(s, reps)

    # Truncate the background signal  
    s = s[:x_len]

    # Check if the lengths are the same
    assert x_len == s.shape[0]

    # Convert SNRdb to linear
    snr = 10**(snr_db/10)
    
    # Estimate noise and signal power
    sp = np.sqrt(np.mean(s**2))
    xp = np.sqrt(np.mean(x**2))

    # Calculate desired noise power based on the SNR value
    sp_target = xp/snr

    # Scale factor noise signal
    alpha = sp_target/sp

    # Sum speech and background noise
    y = x + alpha*s

    # Save file
    sf.write(out_filepath, data=y, samplerate=sr)

def clip_signal(in_filepath, out_filepath, clip_factor='10', sr=16000):
    # Read signal
    x, sr = sf.read(in_filepath)
    
    # Calculate percentiles
    lower_percentile = clip_factor/2
    higher_percentile = 100 - lower_percentile
    percentile_values = np.percentile(x, [lower_percentile, higher_percentile])
    
    # Clip signal
    x[x>percentile_values[1]], x[x<percentile_values[0]] = percentile_values[1], percentile_values[0]
    
    # Store clipped signal
    sf.write(out_filepath, data=x, samplerate=sr)

# UNSEEN degradations
def vorbis(in_filepath, out_filepath, quality='3', sr=16000):
    # Compress vorbis
    vorbis_outpath = os.path.splitext(out_filepath)[0] + '.ogg'
    cmd = f'ffmpeg -y -i {in_filepath} -c:a libvorbis -qscale:a {quality} {vorbis_outpath}'.split(' ')
    subprocess.call(cmd)

    # Convert to wav
    wav_convert = f'ffmpeg -y -i {vorbis_outpath} -ar {sr} {out_filepath}'.split(' ')
    subprocess.call(wav_convert)
    os.remove(vorbis_outpath)

def reverb(in_filepath, out_filepath, p=50, sr=16000):
    wave, _  = torchaudio.load(in_filepath)
    d_wave, _ = torchaudio.sox_effects.apply_effects_tensor(wave, sample_rate=sr, effects=[['reverb', f'{p}']])
    d_wave = (d_wave[0,:] + d_wave[1,:])/2
    sf.write(out_filepath, data=d_wave, samplerate=sr)


import yaml
import os
import subprocess
from degradations import mp3, opus, noise, clip_signal
import pandas as pd
import random
random.seed(0)

with open('src/config/config_audio_degrader.yaml') as file:
    config = yaml.load(file, Loader=yaml.FullLoader)

sr = config['sr']
in_dir = os.path.join(config['root'], config['in_dir_train'])
in_dir_wav = os.path.join(config['root'], config['in_dir_train_wav'])
out_dir = os.path.join(config['root'], config['out_dir_train'])

# Specify the total number of conditions (5 degradations x 4 intensity levels )
n_cond = 20

filepath_deg, filename_deg, filepath_ref, filename_ref = [], [], [], []
abs_filepath_deg, abs_filepath_ref = [], []
for root, dirs, files in os.walk(in_dir):
    for filename in files:
        if (filename.endswith('.flac') | (filename.endswith('.wav'))):  
            in_filepath = os.path.join(root, filename)
            filename = os.path.splitext(filename)[0] + '.wav'
            
            # Store filename ref 
            filename_ref += n_cond * [filename]

            # Store filepath ref (relative path)
            sub_dir = root.split('/')[-2:]
            sub_dir = '/'.join(sub_dir)

            # Make subdirs
            if not os.path.isdir(os.path.join(in_dir_wav, sub_dir)):
                os.makedirs(os.path.join(in_dir_wav, sub_dir))
            in_filepath_wav = os.path.join(in_dir_wav, sub_dir, filename)
            
            # Convert to wav for visqol
            if not in_filepath.endswith('.wav'):
                wav_convert = f'ffmpeg -y -i {in_filepath} -ar {sr} {in_filepath_wav}'.split(' ')
                subprocess.call(wav_convert)

            filepath_ref_relative = os.path.join(sub_dir, filename)
            filepath_ref += n_cond * [filepath_ref_relative]

            # Store absolute filepath ref
            abs_filepath_ref += n_cond * [in_filepath_wav]

            # *** MP3 ****
            # Create out subdir
            subdir_name = 'MP3'
            out_dir_mp3 = os.path.join(out_dir, subdir_name)
            if not os.path.isdir(out_dir_mp3):
                os.makedirs(out_dir_mp3)
            
            # Call mp3 codec for all conditions
            for cond in config['mp3_train']:
                # Build filename deg file
                filename_deg_file = filename.split('.')[0] + f'_{subdir_name}_{cond}.wav'
                filename_deg.append(filename_deg_file)
                
                # Build filepath deg file
                out_filepath = os.path.join(out_dir_mp3, filename_deg_file)
                filepath_deg.append(os.path.join(subdir_name, filename_deg_file))
                abs_filepath_deg.append(out_filepath)

                mp3(in_filepath_wav, out_filepath, bitrate=cond, sr=config['sr'])
                cmd_normalization = f'ffmpeg-normalize {out_filepath} -o {out_filepath} -f -ar {16000}'.split(' ')
                subprocess.call(cmd_normalization)


            # *** OPUS ***
            # Create out subdir
            subdir_name = 'OPUS'
            out_dir_opus = os.path.join(out_dir, subdir_name)
            if not os.path.isdir(out_dir_opus):
                os.makedirs(out_dir_opus)
                    
            # Call mp3 codec for all conditions
            for cond in config['opus_train']:
                filename_deg_file = filename.split('.')[0] + f'_{subdir_name}_{cond}.wav'
                filename_deg.append(filename_deg_file)

                out_filepath = os.path.join(out_dir_opus, filename_deg_file)
                filepath_deg.append(os.path.join(subdir_name, filename_deg_file))
                abs_filepath_deg.append(out_filepath)

                opus(in_filepath_wav, out_filepath, bitrate=cond, sr=config['sr'])
                cmd_normalization = f'ffmpeg-normalize {out_filepath} -o {out_filepath} -f -ar {16000}'.split(' ')
                subprocess.call(cmd_normalization)

            # *** NOISE ***
            # Create out subdir
            subdir_name = 'NOISE'
            out_dir_noise = os.path.join(out_dir, subdir_name)
            if not os.path.isdir(out_dir_noise):
                os.makedirs(out_dir_noise)
                
            # Pick one noise file
            noise_dir = os.path.join(config['root_noise'], config['noise_dir_train'])
            list_noise_files = [nx for nx in os.listdir(noise_dir) if nx.endswith('.wav')]
            noise_path = os.path.join(noise_dir, random.choice(list_noise_files))
            for cond in config['noise_train']:
                str_cond = str(cond)
                filename_deg_file = filename.split('.')[0] + f'_{subdir_name}_{str_cond}.wav'
                filename_deg.append(filename_deg_file)

                out_filepath = os.path.join(out_dir_noise, filename_deg_file)
                filepath_deg.append(os.path.join(subdir_name, filename_deg_file))
                abs_filepath_deg.append(out_filepath)
                
                noise(in_filepath_wav, noise_path, out_filepath, snr_db=cond, sr=config['sr'])
                cmd_normalization = f'ffmpeg-normalize {out_filepath} -o {out_filepath} -f -ar {16000}'.split(' ')
                subprocess.call(cmd_normalization)

            # *** CLIP ***
            # Create out subdir
            subdir_name = 'CLIP'
            out_dir_clip = os.path.join(out_dir, subdir_name)
            if not os.path.isdir(out_dir_clip):
                os.makedirs(out_dir_clip)
            
            # Clip
            for cond in config['clip_train']:
                str_cond = str(cond)
                filename_deg_file = filename.split('.')[0] + f'_{subdir_name}_{str_cond}.wav'
                filename_deg.append(filename_deg_file)

                out_filepath = os.path.join(out_dir_clip, filename_deg_file)
                filepath_deg.append(os.path.join(subdir_name, filename_deg_file))
                abs_filepath_deg.append(out_filepath)            

                clip_signal(in_filepath_wav, out_filepath, clip_factor=cond, sr=config['sr'])
                cmd_normalization = f'ffmpeg-normalize {out_filepath} -o {out_filepath} -f -ar {16000}'.split(' ')
                subprocess.call(cmd_normalization)

# Store data
df = pd.DataFrame({'filename_ref': filename_ref, 'filepath_ref': filepath_ref, 'filename_deg': filename_deg, 'filepath_deg': filepath_deg})
df.to_csv('degraded_data.csv', index=False)

# Store data for ViSQOL
df_visqol = pd.DataFrame({'filepath_ref': abs_filepath_ref, 'filepath_deg': abs_filepath_deg})
df_visqol.to_csv('degraded_data_visqol_format.csv', index=False)


from typing import Any
import torch.nn as nn
import torch
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence, PackedSequence
import math

class TripletModel(nn.Module):
    def __init__(self, ssl_model, ssl_out_dim, emb_dim=256):
        super(TripletModel, self).__init__()
        self.ssl_model = ssl_model
        self.ssl_features = ssl_out_dim
        self.embedding_layer = nn.Sequential(
            nn.ReLU(),
            nn.Linear(self.ssl_features, emb_dim)
        )
    
    def forward(self, wav, lengths=None):
        wav = wav.squeeze(1)  ## [batches, audio_len]
        res = self.ssl_model(wav, mask=False, features_only=True)
        x = res['x']
        x_tr = torch.mean(x, 1)
        x = self.embedding_layer(x_tr)
        x = torch.nn.functional.normalize(x, dim=1)
        return x


class Origw2v(nn.Module):
    def __init__(self, ssl_model, ssl_out_dim):
        super(Origw2v, self).__init__()
        self.ssl_model = ssl_model
        self.ssl_features = ssl_out_dim
    
    def forward(self, wav, lengths=None):
        wav = wav.squeeze(1)  ## [batches, audio_len]
        res = self.ssl_model(wav, mask=False, features_only=True)
        x = res['x']
        x_tr = torch.mean(x, 1)
        #x  = torch.nn.functional.normalize(x, dim=1)
        return x_tr

class MosPredictor(nn.Module): 
    def __init__(self, ssl_model, ssl_out_dim):
        super(MosPredictor, self).__init__()
        self.ssl_model = ssl_model
        self.ssl_features = ssl_out_dim
        self.out_layer = nn.Linear(self.ssl_features, 1)
        
    def forward(self, wav):
        wav = wav.squeeze(1)  ## [batches, audio_len]
        res = self.ssl_model(wav, mask=False, features_only=True)
        x = res['x']
        x = torch.mean(x, 1)
        x = self.out_layer(x)
        return x.squeeze(1)

class MosPredictorTriplet(nn.Module):
    def __init__(self, ssl_model, ssl_out_dim, emb_dim=256):
        super(MosPredictorTriplet, self).__init__()
        self.ssl_model = ssl_model
        self.ssl_features = ssl_out_dim
        self.embedding_layer = nn.Sequential(
            nn.ReLU(),
            nn.Linear(self.ssl_features, emb_dim)
        )        
        self.out_layer = nn.Linear(emb_dim, 1)
        
    def forward(self, wav):
        wav = wav.squeeze(1)  ## [batches, audio_len]
        res = self.ssl_model(wav, mask=False, features_only=True)
        x = res['x']
        x = torch.mean(x, 1)
        x = self.embedding_layer(x)
        x = torch.nn.functional.normalize(x, dim=1)
        x = 1+4*torch.sigmoid(self.out_layer(x))
        return x.squeeze(1)

class FeatureEncoder(nn.Module):
    def __init__(self, cnn_1=16, cnn_2=32, cnn_3=64):
        super().__init__()

        self.dropout = nn.Dropout2d(p=0.2)
        
        # Dim = (1 x 48 x 16)
        self.layer1 = nn.Sequential(
            nn.Conv2d(1, cnn_1, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(cnn_1),
            nn.ReLU(),
        )
        
        # Dim = (16 x 48 x 16)
        self.layer2 = nn.Sequential(
            nn.Conv2d(cnn_1, cnn_2, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(cnn_2),
            nn.ReLU(),
            nn.AdaptiveMaxPool2d(output_size=[24, 8])
        )

        # Dim = (32 x 24 x 8)
        self.layer3 = nn.Sequential(
            nn.Conv2d(cnn_2, cnn_2, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(cnn_2),
            nn.ReLU(),
        )
        
        # Dim = (32 x 24 x 8)
        self.layer4 = nn.Sequential(
            nn.Conv2d(cnn_2, cnn_3, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(cnn_3),
            nn.ReLU(),
            nn.AdaptiveMaxPool2d(output_size=[12, 4])
        )
        
        # Dim = (64 x 12 x 4)
        self.layer5 = nn.Sequential(
            nn.Conv2d(cnn_3, cnn_3, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(cnn_3),
            nn.ReLU(),
            nn.AdaptiveMaxPool2d(output_size=[6, 1])
        )

        self.out_fmap_size = [6, 1]
        self.out_num_filters = cnn_3
    
    def forward(self, x):
        # First block
        x = self.layer1(x)
        x = self.dropout(x)
        x = self.layer2(x)
        
        # Second block
        x = self.layer3(x)
        x = self.dropout(x)
        x = self.layer4(x)

        # Third block
        x = self.layer5(x)

        # Flatten (384, 1)
        x = x.view(-1, self.out_fmap_size[0] * self.out_fmap_size[1] * self.out_num_filters)
        return x

class TimeModeling(nn.Module):
    def __init__(self, input_size=384, hidden_size=128, num_layers=1, batch_first=True, bidirectional=True):
        super().__init__()
        self.lstm = nn.LSTM(input_size, hidden_size, num_layers, batch_first, bidirectional=bidirectional)
    
    def forward(self, x, lengths):
        x = pack_padded_sequence(x, lengths, batch_first=True, enforce_sorted=False)             
        self.lstm.flatten_parameters()
        x = self.lstm(x)[0]
        x, _ = pad_packed_sequence(x, batch_first=True, padding_value=0.0, total_length=int(lengths.max().item()))          
        # Dim = (Batch size, Num sequences, 2*hidden_size)
        return x

class LSTMPooling(nn.Module):
    def __init__(self):
        super().__init__()
    
    def forward(self, x, lengths):
        # x = (Batch size, Num seq, Hidden size)
        bs = x.shape[0]
        hidden_size = x.shape[-1]
        new_size = 2
        max_len = int(lengths.max().item())

        # Split forward and backward LSTM into a new dimension
        x = x.view(bs, max_len, new_size, hidden_size//2)

        # Extract the last valid sequence for forward LSTM
        lstm_forward = x[torch.arange(x.shape[0]), lengths.type(torch.long)-1, 0, :]

        # Extract first sequence for backward (which corresponds to the last hidden state)
        lstm_backward = x[:,0,1,:]

        # Get output feature vector
        x = torch.cat([lstm_forward, lstm_backward], dim=1)

        return x

class SpecgramModel(nn.Module):
    def __init__(self, cnn_1=16, cnn_2=32, cnn_3=64, input_size=384, hidden_size=256, num_layers=1, batch_first=True, bidirectional=True, emb_dim=256):
        super().__init__()
        self.feature_encoder = FeatureEncoder(cnn_1, cnn_2, cnn_3)
        self.time_model = TimeModeling(input_size, hidden_size, num_layers, batch_first, bidirectional)
        self.time_pooling = LSTMPooling()
        self.embedding_layer = nn.Sequential(
            #nn.ReLU(), 
            nn.Linear(hidden_size*2, emb_dim)
        )

    def forward(self, x, lengths):
        # Pack padded sequence
        lengths = lengths.cpu()
        batch_max_len = int(lengths.max().item())
        x_packed = pack_padded_sequence(x, lengths, batch_first=True, enforce_sorted=False)
        
        # Cnn encoder
        x = self.feature_encoder(x_packed.data)

        # Pad packed sequence
        x_packed = PackedSequence(x, batch_sizes=x_packed.batch_sizes, sorted_indices=x_packed.sorted_indices, unsorted_indices=x_packed.unsorted_indices)
        x, _ = pad_packed_sequence(x_packed, batch_first=True, total_length=batch_max_len)        

        # LSTM
        x = self.time_model(x, lengths)

        # Pool last hidden state forward and first hidden state for backward
        x = self.time_pooling(x, lengths)

        # Get embedding layer for triplet
        x = self.embedding_layer(x)

        # L2 normalization
        x = torch.nn.functional.normalize(x, dim=1)
        return x

# # Test feature extractor
#path = '/media/alergn/hdd/datasets/audio/speech/LibriSpeech/train-clean-100-wav/911/130578/911-130578-0000.wav'

# # Load audio
# import torchaudio
# wav, _ = torchaudio.load(path)
# melspec = torchaudio.transforms.MelSpectrogram(n_mels=48)(wav)
# log_mel = torch.log10(melspec)
# patches = log_mel.unfold(dimension=2, size=16, step=8).permute(2, 0, 1, 3)

# #wav = torchaudio.transforms.Resample(new_freq=8000)(wav)

# # Create feature extractor
# #conv_layers = [(512, 10, 5), (512, 3, 2), (512, 3, 2), (512, 3, 2), (512, 3, 2), (512, 2, 2), (512, 2, 2)]
# model = SpecgrModel()
# x = torch.cat([patches.unsqueeze(0), patches.unsqueeze(0)])
# x = model(x, torch.Tensor([139, 100]))

# ************ MODELS THAT DIDN'T SHOW ANY IMPROVEMENTS BELOW ***********
# ***********************************************************************

class QualityEmbeddings(nn.Module):
    def __init__(self, context_dim=512, emb_dim=256):
        super().__init__()

        self.embeddings = nn.Sequential(
            nn.ReLU(),
            nn.Linear(context_dim, emb_dim)
        )
    
    def forward(self, x):
        x = self.embeddings(x)
        x = torch.nn.functional.normalize(x, dim=1)
        return x

class MosPredictorTriplet2(nn.Module):
    def __init__(self, ssl_model, ssl_out_dim, emb_dim=256):
        super(MosPredictorTriplet2, self).__init__()
        self.ssl_model = ssl_model
        self.ssl_features = ssl_out_dim
        self.embedding_layer = QualityEmbeddings2(ssl_out_dim, emb_dim)
        self.out_layer = nn.Linear(emb_dim, 1)
        
    def forward(self, wav):
        wav = wav.squeeze(1)  ## [batches, audio_len]
        res = self.ssl_model(wav, mask=False, features_only=True)
        x = res['x']
        x = torch.mean(x, 1)
        x = self.embedding_layer(x)
        x = torch.nn.functional.normalize(x, dim=1)
        x = 1+4*torch.sigmoid(self.out_layer(x))
        return x.squeeze(1)


class QualityEmbeddings2(nn.Module):
    def __init__(self, context_dim=512, emb_dim=256):
        super().__init__()

        self.embeddings = nn.Sequential(
            nn.Linear(context_dim, 512),
            nn.BatchNorm1d(512),
            nn.ReLU(),

            nn.Linear(512, 512),
            nn.BatchNorm1d(512),
            nn.ReLU(),

            nn.Linear(512, emb_dim),
        )
    
    def forward(self, x):
        x = self.embeddings(x)
        return x    

class QualityNet(nn.Module):
    def __init__(self, ssl_model, ssl_out_dim, emb_dim=256):
        super().__init__()
        self.ssl_model = ssl_model
        self.ssl_features = ssl_out_dim
        self.embedding_layer = nn.Sequential(
            nn.ReLU(),
            nn.Linear(self.ssl_features, emb_dim)
        )

        self.mos_net = nn.Sequential(
            nn.Linear(emb_dim, 128),
            nn.BatchNorm1d(128),
            nn.ReLU(),
            #nn.Dropout(),

            nn.Linear(128, 64),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            #nn.Dropout(),

            nn.Linear(64, 32),
            nn.BatchNorm1d(32),
            nn.ReLU(),
            #nn.Dropout(),

            # MOS prediction
            nn.Linear(32, 1)
        )
        
    def forward(self, wav):
        wav = wav.squeeze(1)  ## [batches, audio_len]
        res = self.ssl_model(wav, mask=False, features_only=True)
        x = res['x']
        x = torch.mean(x, 1)
        x = self.embedding_layer(x)
        x = torch.nn.functional.normalize(x, dim=1)
        x = self.mos_net(x)
        return x.squeeze(1)
    
class PairsModel(nn.Module):
    def __init__(self, ssl_model, ssl_out_dim, emb_dim=256):
        super(PairsModel, self).__init__()
        self.ssl_model = ssl_model
        self.ssl_features = ssl_out_dim
        self.embedding_layer = nn.Linear(self.ssl_features, 1)
        
    def forward(self, wav):
        wav = wav.squeeze(1)  ## [batches, audio_len]
        res = self.ssl_model(wav, mask=False, features_only=True)
        x = res['x']
        x = torch.mean(x, 1)
        x = self.embedding_layer(x)
        #x = torch.nn.functional.normalize(x, dim=1)
        return x.squeeze(1)

class MultiTaskModel(nn.Module):
    def __init__(self, ssl_model, ssl_out_dim):
        super(MultiTaskModel, self).__init__()
        self.ssl_model = ssl_model
        self.ssl_features = ssl_out_dim
        self.embedding_triplet = nn.Linear(4*ssl_out_dim, 256)
        self.out_layer = nn.Linear(self.ssl_features + 256, 1)
        
    def forward(self, wav):
        wav = wav.squeeze(1)  ## [batches, audio_len]
        res = self.ssl_model(wav, mask=False, features_only=True)
        x = res['x']

        # Features from 6 to 10 layers (where is phonetic informaton)
        middle_layers = torch.cat([ torch.mean(tr_layers[0], dim=0) for tr_layers in res['layer_results'][0:3]], 1)
        triplet_features = self.embedding_triplet(torch.nn.functional.relu(middle_layers)) 
        triplet_features = torch.nn.functional.normalize(triplet_features)

        tr_avg_pool = torch.mean(x, 1)
        out_features = torch.cat([tr_avg_pool, triplet_features], dim=1)
        x = self.out_layer(out_features)

        return x.squeeze(1), triplet_features

class MultiTaskPairModel(nn.Module):
    def __init__(self, ssl_model, ssl_out_dim):
        super(MultiTaskPairModel, self).__init__()
        self.ssl_model = ssl_model
        self.ssl_features = ssl_out_dim
        self.mos_layer = nn.Linear(self.ssl_features , 1)
        self.rank_layer = nn.Linear(self.ssl_features, 1)
        
    def forward(self, wav):
        wav = wav.squeeze(1)  ## [batches, audio_len]
        res = self.ssl_model(wav, mask=False, features_only=True)
        #x = res['x']
        # Take first transformer layer for multi task learning
        x = res['layer_results'][1][0].permute(1, 0, 2)

        # Context representations 
        tr_avg_pool = torch.mean(x, 1)

        # Multi task output layers mos, rank
        mos_p = self.mos_layer(tr_avg_pool)
        rank_p = self.rank_layer(tr_avg_pool)
        return mos_p.squeeze(1), rank_p.squeeze(1)

class MultiTaskPairModelBce(nn.Module):
    def __init__(self, ssl_model, ssl_out_dim):
        super(MultiTaskPairModelBce, self).__init__()
        self.ssl_model = ssl_model
        self.ssl_features = ssl_out_dim
        self.out_layer = nn.Linear(self.ssl_features , 1)
        
    def forward(self, wav):
        wav = wav.squeeze(1)  ## [batches, audio_len]
        res = self.ssl_model(wav, mask=False, features_only=True)
        x = res['x']

        # Context representations 
        tr_avg_pool = torch.mean(x, 1)

        # Multi task output layers mos, rank
        mos_p = self.out_layer(tr_avg_pool)
        return mos_p.squeeze(1)


class CnnEncoder(nn.Module):
    def __init__(self, cnn_1=16, cnn_2=32, cnn_3=64):
        super().__init__()

        self.dropout = nn.Dropout2d(p=0.2)

        self.layer1 = nn.Sequential(
            nn.Conv2d(1, cnn_1, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(cnn_1),
            nn.ReLU(),
            nn.AdaptiveMaxPool2d(output_size=[24, 7])
        )

        self.layer2 = nn.Sequential(
            nn.Conv2d(cnn_1, cnn_2, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(cnn_2),
            nn.ReLU(),
            nn.AdaptiveMaxPool2d(output_size=[12, 5])

        )

        self.layer3 = nn.Sequential(
            nn.Conv2d(cnn_2, cnn_3, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(cnn_3),
            nn.ReLU(),
        )
        
        self.layer4 = nn.Sequential(
            nn.Conv2d(cnn_3, cnn_3, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(cnn_3),
            nn.ReLU(),
            nn.AdaptiveMaxPool2d(output_size=[6, 3])
        )

        self.layer5 = nn.Sequential(
            nn.Conv2d(cnn_3, cnn_3, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(cnn_3),
            nn.ReLU()
        )

        self.layer6 = nn.Sequential(
            nn.Conv2d(cnn_3, cnn_3, kernel_size=3, stride=1, padding=(1, 0)),
            nn.BatchNorm2d(cnn_3),
            nn.ReLU()            
        )
        self.out_fmap_size = [6, 1]
        self.out_num_filters = cnn_3
    
    def forward(self, x):
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.dropout(self.layer3(x))
        x = self.dropout(self.layer4(x))
        x = self.dropout(self.layer5(x))
        x = self.layer6(x)

        # Flatten 
        x = x.view(-1, self.out_fmap_size[0] * self.out_fmap_size[1] * self.out_num_filters)
        return x

class FrameWiseMOS(nn.Module):
    def __init__(self):
        super().__init__()
        self.cnn_enc = CnnEncoder()
        self.fc_mp = nn.Linear(384, 1)
    
    def forward(self, x, lengths):
        # Pack padded sequence
        batch_max_len = x.shape[1]
        x_packed = pack_padded_sequence(x, lengths, batch_first=True, enforce_sorted=False)
        
        # Cnn encoder
        x = self.cnn_enc(x_packed.data)

        # Pad packed sequence
        x_packed = PackedSequence(x, batch_sizes=x_packed.batch_sizes, sorted_indices=x_packed.sorted_indices, unsorted_indices=x_packed.unsorted_indices)
        x, _ = pad_packed_sequence(x_packed, batch_first=True, total_length=batch_max_len)        

        # Extract one sequence (torch.max acts as a selector and backpropagation will only work on the extracted sequence)
        mp_features = torch.max(x, 1)[0]
        x = self.fc_mp(mp_features)
        return x.squeeze(1)

class FrameWiseMOSmtl(nn.Module):
    def __init__(self):
        super().__init__()
        self.cnn_enc = CnnEncoder()
        self.mos_layer = nn.Linear(384, 1)
        self.rank_layer = nn.Linear(384, 1)
    
    def forward(self, x, lengths):
        # Pack padded sequence
        batch_max_len = x.shape[1]
        x_packed = pack_padded_sequence(x, lengths, batch_first=True, enforce_sorted=False)
        
        # Cnn encoder
        x = self.cnn_enc(x_packed.data)

        # Pad packed sequence
        x_packed = PackedSequence(x, batch_sizes=x_packed.batch_sizes, sorted_indices=x_packed.sorted_indices, unsorted_indices=x_packed.unsorted_indices)
        x, _ = pad_packed_sequence(x_packed, batch_first=True, total_length=batch_max_len)        

        # Extract one sequence (torch.max acts as a selector and backpropagation will only work on the extracted sequence)
        mp_features = torch.max(x, 1)[0]
        
        mos = self.mos_layer(mp_features)
        rank = self.rank_layer(mp_features)
        return mos.squeeze(1), rank.squeeze(1)

class Transpose(nn.Module):
    def __init__(self, transpose_dim=-2):
        super().__init__()
        self.transpose_dim = transpose_dim
    
    def forward(self, x):
        return x.transpose(self.transpose_dim, -1)

class FeatureExtractor2(nn.Module):
    def __init__(self):
        super().__init__()
        
        self.num_layers = 20
        self.num_filters = 16
        self.filter_size = 15
        self.cnn_encoder = nn.ModuleList()
        self.batchnorm_encoder = nn.ModuleList()

        self.in_ch =  [1] + [(i + 1) * self.num_filters for i in range(self.num_layers-1)]
        self.out_ch = [(i + 1) * self.num_filters for i in range(self.num_layers)]
        
        nchan = self.num_filters
        for i in range(self.num_layers):
                    if i==0:
                        chin = 1
                    else:
                        chin = nchan
                    if (i+1)%4==0:
                        nchan = nchan*2
                    self.cnn_encoder.append(nn.Conv1d(chin,nchan,self.filter_size,padding=self.filter_size//2))
                    self.batchnorm_encoder.append(nn.BatchNorm1d(nchan))

    def forward(self, x):
        input = x
        
        for i in range(self.num_layers):
            x = self.cnn_encoder[i](x)
            x = self.batchnorm_encoder[i](x)
            x = torch.nn.functional.leaky_relu(x,0.1)
            if (i+1)%4==0:
                x = x[:,:,::2]
        
        x = torch.sum(x,dim=(2))/x.shape[2] # average by channel dimension
        
        return x

class FeatureExtractor(nn.Module):
    # conv_layers will be a list of tuples where each tuple includes (n_ch, k, stride)
    def __init__(self, conv_layers, dropout=0.2):
        super().__init__()

        def init_conv(n_in, n_out, k, stride):
            conv_layer = nn.Conv1d(n_in, n_out, k, stride=stride)
            nn.init.kaiming_normal_(conv_layer.weight)
            return conv_layer        

        def conv_block(n_in, n_out, k, stride, layer_norm=False):
            block = nn.Sequential(
                init_conv(n_in, n_out, k, stride),
                nn.Dropout(p=dropout),
                #nn.Sequential(
                #    Transpose(), 
                #    nn.LayerNorm(n_ch, elementwise_affine=True),
                #    Transpose()
                #),
                nn.ReLU()
            )
            return block
        
        # Create convolutional blocks
        self.conv_layers = nn.ModuleList()
        
        # First layer 1 channel
        n_in = 1
        
        for id_conv_layer, settings_conv_layer in enumerate(conv_layers):
            # Extract settings
            (n_ch, k, stride) = settings_conv_layer

            # Add blocks to ModuleList
            self.conv_layers.append(
                conv_block(n_in, n_ch, k, stride)
            )

            # Update input channels for next layer
            n_in = n_ch
    
    def forward(self, x):   
        # Add channel dimension to mono signal
        #x = x.unsqueeze(1)

        # Pass conv layers
        for conv in self.conv_layers:
            x = conv(x)
        
        return x

class PositionalEncoding(nn.Module):
    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model))
        pe = torch.zeros(max_len, 1, d_model)
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)
        self.register_buffer('pe', pe)

    def forward(self, x):
        """
        Arguments:
            x: Tensor, shape ``[seq_len, batch_size, embedding_dim]``
        """
        x = x + self.pe[:x.size(0)]
        return self.dropout(x)

class ContextEncoder(nn.Module):

    def __init__(self,  d_model: int, nhead: int, d_hid: int,
                 nlayers: int, dropout: float = 0.5):
        super().__init__()
        self.model_type = 'Transformer'
        self.pos_encoder = PositionalEncoding(d_model, dropout)
        encoder_layers = nn.TransformerEncoderLayer(d_model, nhead, d_hid, dropout)
        self.transformer_encoder = nn.TransformerEncoder(encoder_layers, nlayers)
        self.d_model = d_model

    #     self.init_weights()

    # def init_weights(self):
    #     initrange = 0.1
    #     self.linear.bias.data.zero_()
    #     self.linear.weight.data.uniform_(-initrange, initrange)

    def forward(self, src):
        """
        Arguments:
            src: Tensor, shape ``[seq_len, batch_size]``

        Returns:
            output Tensor of shape ``[seq_len, batch_size, ntoken]``
        """
        #src = self.embedding(src) * math.sqrt(self.d_model)
        src = self.pos_encoder(src)
        output = self.transformer_encoder(src)
        return output

class TripletModel2(nn.Module):
    def __init__(self, ssl_model, ssl_out_dim, emb_dim=256):
        super(TripletModel2, self).__init__()
        self.ssl_model = ssl_model
        self.ssl_features = ssl_out_dim
        self.embedding_layer = QualityEmbeddings2(context_dim=ssl_out_dim, emb_dim=emb_dim)
    
    def forward(self, wav, lengths=None):
        wav = wav.squeeze(1)  ## [batches, audio_len]
        res = self.ssl_model(wav, mask=False, features_only=True)
        x = res['x']
        x_tr = torch.mean(x, 1)
        x = self.embedding_layer(x_tr)
        x = torch.nn.functional.normalize(x, dim=1)
        return x

class MyModel(nn.Module):
    def __init__(self, conv_layers, d_model=512, nhead=8, d_hid=3072, nlayers=6, emb_dim=256):
        super().__init__()

        self.feature_encoder = FeatureExtractor(conv_layers)
        self.context_encoder = ContextEncoder(d_model, nhead, d_hid, nlayers)
        self.quality_embeddings = QualityEmbeddings(d_model, emb_dim)

    def forward(self, x):
        # Feature extraction CNN
        x = self.feature_encoder(x)

        # Context representations Transformer
        x = self.context_encoder(x.permute(0, 2, 1))
        x = x.permute(0, 2, 1)

        # Remove time dimension with the mean
        x = torch.mean(x, dim=2)

        # Embedding vec
        x = self.quality_embeddings(x)
        return x

# This architecture uses the backbone feature extractor of w2v and train a transformer from scratch with embedding layer
class SQUAD(nn.Module):
    def __init__(self, ssl_model, d_model=512, nhead=8, d_hid=3072, nlayers=4, dropout=0.5, emb_dim=256):
        super().__init__()
        self.ssl_model = ssl_model
        self.context_encoder = ContextEncoder(d_model, nhead, d_hid, nlayers, dropout)
        self.quality_embeddings = QualityEmbeddings(d_model, emb_dim)
    
    def forward(self, x):
        # Feature extraction pre-trained CNN
        x = x.squeeze(1)  ## [batches, audio_len]
        ssl_features = self.ssl_model(x, mask=False, features_only=True)
        x = ssl_features['features']        
        
        # Context encoder transformer (from scratch), take the mean to remove time dimension
        x = self.context_encoder(x)
        x = torch.mean(x, 1)

        # Quality embeddings
        x = self.quality_embeddings(x)

        return x

class SQUAD2(nn.Module):
    def __init__(self, conv_layers, d_model=512, emb_dim=256):
        super().__init__()
        #self.cnn_backbone = FeatureExtractor(conv_layers)
        self.cnn_backbone = FeatureExtractor2()
        self.quality_embeddings = QualityEmbeddings2(d_model, emb_dim)
    

    def forward(self, x):
        # Feature extraction pre-trained CNN
        #x = x.squeeze(1)  ## [batches, audio_len]
        if len(x.shape) == 2:
            x = x.unsqueeze(1)
        x = self.cnn_backbone(x)       
        
        # Pooling over time dimension (get audio embeddings)
        # x = torch.mean(x, 2)

        # Quality embeddings net
        x = self.quality_embeddings(x)

        # Normalize row-wise each embedding vector
        x =  torch.nn.functional.normalize(x, dim=1)
        return x


import torch
import torch.nn as nn
import fairseq
import yaml
from src.dataloader.triplet_dataloader import TripletDataset, load_processing
from src.models.networks import TripletModel, Origw2v, SpecgramModel
from tqdm import tqdm
import numpy as np
from torch.utils.data import DataLoader
import os
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from datetime import datetime
import random
from scipy.stats import spearmanr, pearsonr
from scipy.optimize import curve_fit
import torch.optim.lr_scheduler as lr_scheduler
import torchaudio
from scipy.spatial.distance import cdist
sns.set_style('darkgrid')

# REPRODUCIBILITY 
SEED = 0
torch.manual_seed(SEED)
np.random.seed(SEED)
random.seed(SEED)
def seed_worker(worker_id):
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)

g = torch.Generator()
g.manual_seed(SEED)
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False

class Training():
    def __init__(self, config_file):
    
        # Configuration file loading
        with open(config_file) as file:
            self.config = yaml.load(file, Loader=yaml.FullLoader)

        # Find device
        if torch.cuda.is_available:
            self.DEVICE = 'cuda'
        else:
            self.DEVICE = 'cpu'
        print(f'Device: {self.DEVICE}')

        if self.config['architecture'] == 'w2v': 
            # Load SSL model if using wav2vec
            CHECKPOINT_PATH = self.config['checkpoint_path']
            SSL_OUT_DIM = self.config['ssl_out_dim']
            EMB_DIM = self.config['emb_dim']

            w2v_model, _, _ = fairseq.checkpoint_utils.load_model_ensemble_and_task([CHECKPOINT_PATH])
            ssl_model = w2v_model[0] 
            ssl_model.remove_pretraining_modules()

            if self.config['eval_w2v']:
                self.model = Origw2v(ssl_model, SSL_OUT_DIM)
            else:
                self.model = TripletModel(ssl_model, SSL_OUT_DIM, EMB_DIM)
            self.model.to(self.DEVICE)

            # Choose if you want to 1) Freeze only ConvNet 2) Freeze ConvNet + Transformer 3) Finetune the entire network
            # Freeze only ConvNet
            if self.config['experiment_name'] == 'Training':
                if self.config['freeze_convnet']:
                    self.model.ssl_model.feature_extractor.requires_grad_(False)
                
                # Freeze both ConvNet and Transformer (no finetuning)
                if self.config['freeze_all']:
                    self.model.ssl_model.feature_extractor.requires_grad_(False)
                    self.model.ssl_model.encoder.requires_grad_(False)
        
        # Specgram model (results not reported in paper)
        elif self.config['architecture'] == 'SpecgramModel':
            self.model = SpecgramModel()
            self.model.to(self.DEVICE)
        
        if self.config['experiment_name'] == 'Training':
            # Create dataloaders
            self.current_level = self.config['current_level']
            self.train_set = TripletDataset(self.config, data_mode='train_df', level=self.current_level)
            if self.config['architecture'] == 'w2v':
                collate_fn = self.train_set.collate_fn
            else:
                collate_fn = None
            self.train_loader = DataLoader(self.train_set, batch_size=self.config['train_bs'], shuffle=True, num_workers=self.config['num_workers'], collate_fn=collate_fn)
            self.valid_set = TripletDataset(self.config, data_mode='valid_df', level=self.current_level)
            self.valid_loader = DataLoader(self.valid_set, batch_size=self.config['val_bs'], shuffle=False, num_workers=self.config['num_workers'], collate_fn=collate_fn)

            # Create loss
            self.criterion = nn.TripletMarginLoss(margin=self.config['margin'])
            
            # Create optimizer
            self.optim = torch.optim.Adam(self.model.parameters(), lr=self.config['lr'])

            # Create optimizer with adaptive learning rate
            if self.config['architecture'] == 'w2v':
                if self.config['freeze_convnet']:
                    params_names_embeddings = [f'embeddings.{j}.weight' for j in range(7)] + [f'embeddings.{j}.bias' for j in range(7)]
                    params_pt = [param for name, param in self.model.named_parameters() if name not in params_names_embeddings]
                    params_embeddings = [param for name, param in self.model.named_parameters() if name in params_names_embeddings]                 
                    # Overwrite optimizer
                    self.optim = torch.optim.Adam([
                        {'params': params_pt, 'lr': 1e-5},
                        {'params': params_embeddings}
                    ], lr=self.config['lr'])
        
            # Create learning rate scheduler
            self.lr_scheduler = lr_scheduler.ExponentialLR(self.optim, gamma=self.config['lr_decay_factor'])

    def train(self, model, dataloader, optimizer, criterion):
        model.train()
        total_loss = 0.0

        # Batch data (Anchor, Positive, Negative)
        for batch_index, (A, P, N, A_lengths, P_lengths, N_lengths) in enumerate(tqdm(dataloader)):

            A, P, N, A_lengths, P_lengths, N_lengths = A.to(self.DEVICE), P.to(self.DEVICE), N.to(self.DEVICE), A_lengths, P_lengths, N_lengths

            A_embs = model(A, A_lengths)
            P_embs = model(P, P_lengths)
            N_embs = model(N, N_lengths)

            loss = criterion(A_embs, P_embs, N_embs)

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            total_loss += loss.item()

        return total_loss / len(dataloader)

    def eval(self, model, dataloader, criterion):
        model.eval()
        total_loss = 0.0
        
        with torch.no_grad():
            for batch_index, (A, P, N, A_lengths, P_lengths, N_lengths) in enumerate(tqdm(dataloader)):

                A, P, N, A_lengths, P_lengths, N_lengths = A.to(self.DEVICE), P.to(self.DEVICE), N.to(self.DEVICE), A_lengths, P_lengths, N_lengths

                A_embs = model(A, A_lengths)
                P_embs = model(P, P_lengths)
                N_embs = model(N, N_lengths)

                loss = criterion(A_embs, P_embs, N_embs)

                total_loss += loss.item()

        return total_loss / len(dataloader)

    def training_loop(self):
        # Start Training Loop
        now = datetime.now()
        dt_string = now.strftime("%d-%m-%Y_%H-%M-%S")
        self.PATH_DIR = os.path.join('out_models', self.config['out_dir'], dt_string)
        if not os.path.isdir(self.PATH_DIR):
            os.makedirs(self.PATH_DIR)

        # Save configuration file used
        dump_path_config = os.path.join(self.PATH_DIR, 'config.yaml')
        with open(f"{dump_path_config}", 'w') as file:
            out_config_file = yaml.dump(self.config, file)

        best_valid_loss = np.Inf
        EPOCHS = self.config['num_epochs']

        counter = 0
        for i in range(EPOCHS):
            train_loss = self.train(self.model, self.train_loader, self.optim, self.criterion)
            valid_loss = self.eval(self.model, self.valid_loader, self.criterion)

            if valid_loss < best_valid_loss:
                path_best_model = os.path.join(self.PATH_DIR, 'best_model.pt')
                torch.save(self.model.state_dict(), path_best_model)
                best_valid_loss = valid_loss
                print("Saved Weights Success")
                counter = 0
            else:
                counter += 1
            
            # Check learning rate update
            if (counter + 1) % self.config['lr_decay_step'] == 0:
                self.lr_scheduler.step()

            print(f"COUNTER:  {counter}/{self.config['patience']}")
            print(f'LR: {self.lr_scheduler.get_last_lr()}')
            
            #print(f'CURRENT LEVEL: {self.current_level}')

            if counter > self.config['patience']:
                print('Stop training, counter greater than patience')
                break
            
            print(f"EPOCHS: {i+1} train_loss : {train_loss}")
            print(f"EPOCHS: {i+1} valid_loss : {valid_loss}")
            print('\n')
    
    # Function that extract NOMAD embeddings of a db and store them in a csv
    def get_embeddings_csv(self, model, file_names, root=False):
        file_names_arr = np.array(file_names)
        embeddings = []

        model.eval()
        with torch.no_grad():
            for i, filename_anchor in enumerate(tqdm(file_names_arr)):
                if root:
                    filepath = os.path.join(root, filename_anchor)
                else:
                    filepath = filename_anchor
                
                test_speech = load_processing(filepath, trim=False)
                lengths = None
                
                # Get log mel spectrograms
                if self.config['architecture'] == 'SpecgramModel':
                    test_speech = torchaudio.transforms.MelSpectrogram(
                    sample_rate=self.config['sampling_rate'],
                    n_fft=self.config['n_fft'],
                    win_length=self.config['win_length'],
                    hop_length=self.config['hop_length'],
                    n_mels=self.config['n_mels']
                )(test_speech)
                    test_speech = torch.log10(test_speech + np.finfo(float).eps)

                    # Split into patches
                    test_speech = test_speech.unfold(dimension=2, size=self.config['patch_length'], step=self.config['patch_length']//2).permute(2, 0, 1, 3)
                    lengths = test_speech.shape[0]

                    # Store lengths into a tensor 
                    if not torch.is_tensor(lengths):
                        lengths = torch.Tensor([lengths])
                    
                    # Add batch size dimension (always equal to 1 during test)
                    test_speech.unsqueeze_(0)
                
                test_speech = test_speech.to(self.DEVICE)
                anc_embeddings = model(test_speech, lengths)
                embeddings.append(anc_embeddings.squeeze().cpu().detach().numpy())

            embeddings = np.array(embeddings)
            embeddings = pd.DataFrame(embeddings)
            df_emb = pd.concat([file_names.reset_index(), embeddings], axis=1).drop('index', axis=1)
        
        return df_emb

    def order_three(self, x, a, b, c, d):
        return a*x + b*x**2 + c*x**3 + d

    # Evaluate audio quality with non-matching references
    def eval_audio_quality(self, model_path):
        # Load model weights
        if not self.config['eval_w2v']:
            self.model.load_state_dict(torch.load(model_path))  
        self.model.eval()

        # Get dataframe of the test set
        test_data = pd.read_csv(self.config['test_db_file'])

        # Filter db you want to test, use None to use every db
        if self.config['db'] is not None:
            test_data = test_data[test_data['db'].isin(self.config['db'])]

        # Filter conditions you want to test, use None to use every condition
        db_test = self.config['db']
        if self.config['conds'] is not None:
            conds = self.config['conds']
            print(f'Testing DB: {db_test}, conds: {conds}')
            test_data = test_data[test_data['condition'].str.contains('|'.join(conds))]

        # Get reference files from a clean set (paper uses TSP speech database)
        ref_embeddings = self.get_nmr_embeddings()
        ref_embeddings.set_index('reference', inplace=True)

        # Loop over each database
        db_groups = test_data.groupby('db')

        for db_name, db in db_groups:
            print(db_name)

            # Get test embeddings
            df_emb = self.get_embeddings_csv(self.model, db['filepath_deg'], root=self.config['test_root_wav'])
            test_embeddings = df_emb.set_index('filepath_deg')
            test_names = df_emb.merge(db, on='filepath_deg')[['filepath_deg', 'condition', 'mos']]         
            
            # Calculate distance of each pair then take the average over all the non-matching references for each test audio
            euclid_dist = cdist(test_embeddings, ref_embeddings)
            avg_dist_nmr = np.mean(euclid_dist, axis=1)
            names = test_embeddings.index

            # Add distances 
            df_dist = pd.DataFrame({'filepath_deg': names, 'Distance': avg_dist_nmr})
            df_dist = df_dist.merge(test_names, on='filepath_deg')
            df_dist = df_dist.groupby('condition').mean()

            # Compute third order poly mapping (performance in the paper are reported without mapping)
            #df_grouped = df_dist.groupby('condition')['mos', 'Distance'].mean()
            popt3, _ = curve_fit(self.order_three, df_dist['Distance'].values, df_dist['mos'].values)
            a3, b3, c3, d3 = popt3
            df_dist['Distance_map'] = df_dist['Distance'].apply(lambda x: self.order_three(x,a3,b3,c3,d3))

            # Scatter plot MOS - Embedding Distances
            sns.scatterplot(data=df_dist, x='mos', y='Distance_map')
            plt.xlabel('Actual MOS')
            plt.ylabel(f'Dist w.r.t. clean embeddings')
            plt.xlim([1, 5])
            plt.ylim([1, 5])
            plt.tight_layout()
            out_dir = '/'.join(self.config['nomad_model_path'].split('/')[:-1])
            plt.savefig(os.path.join(out_dir, f'{db_name}_embeddings.png'))
            plt.close()
            
            # Performance Evaluation Spearman
            SRCC, _ = spearmanr(df_dist['Distance'], df_dist['mos'])
            print(f'SRCC: {np.round(SRCC, 2)}')
            SRCC, _ = spearmanr(df_dist['Distance_map'], df_dist['mos'])
            print(f'SRCC 3rd map: {np.round(SRCC, 2)}')

            # Performance Evaluation Pearson
            PCC, _ = pearsonr(df_dist['Distance'], df_dist['mos'])
            print(f'PCC: {np.round(PCC, 2)}')
            PCC, _ = pearsonr(df_dist['Distance_map'], df_dist['mos'])
            print(f'PCC 3rd map: {np.round(PCC, 2)}')
        
    def eval_degr_level(self, model_path):
        # Load model weights
        self.model.load_state_dict(torch.load(model_path))
        self.model.eval()
        
        # Get validation set embeddings
        valid_set = TripletDataset(self.config, data_mode='valid_df', level=self.config['current_level'])
        df_emb = self.get_embeddings_csv(self.model, valid_set.dataset['Anchor'], root=self.config['root'])

        # Get non-matching reference embeddings
        ref_embeddings = self.get_nmr_embeddings()

        # Get validation set anchors - test embeddings
        test_embeddings = df_emb.iloc[:, 1:].to_numpy()
        anc_names = df_emb['Anchor']

        # Take pairwise euclidean distance and then average over non-matching references
        euclid_dist = cdist(test_embeddings, ref_embeddings.iloc[:,1:].to_numpy())
        avg_dist_nmr = np.mean(euclid_dist, axis=1)

        # DF Store 
        df_dist = pd.DataFrame({'Anchor': anc_names, 'Distance': avg_dist_nmr})
        df_dist.sort_values(by='Distance', inplace=True)

        # Boxplot validation set sorted 
        plt.figure(figsize=(50, 20))
        sns.set(font_scale=8.0)
        df_dist['condition'] = [x.split('_')[1] + ' ' + x.split('_')[2].split('.')[0] for x in df_dist['Anchor']]
        order = df_dist.groupby('condition')['Distance'].mean().sort_values().index
        sns.boxplot(df_dist, x='condition', y='Distance', order=order, showmeans=True, meanprops={"markerfacecolor":"white", "markeredgecolor":"blue", "markersize": "50"}, width=0.8, linewidth=6)
        
        # Figure settings
        plt.xticks(rotation=65)
        plt.ylabel(f'NOMAD')
        plt.xlabel('Condition')
        plt.tight_layout()
        out_dir = '/'.join(self.config['nomad_model_path'].split('/')[:-1])
        plt.savefig(os.path.join(out_dir, f'validset_embeddings.png'))
    
    def eval_degradation_intensity(self, model_path):
        if not self.config['eval_w2v']:
            self.model.load_state_dict(torch.load(model_path))  
        self.model.eval()

        # Get non matching references
        ref_embeddings = self.get_nmr_embeddings()
        ref_embeddings.set_index('reference', inplace=True)

        # Get test data
        test_data = pd.read_csv(self.config['test_mono_data'])

        # Loop over each degradation
        test_data_group = test_data.groupby('Degradation')
        
        # Store for scatterplot
        degr_names = []
        degr_data = []
        srcc_scores = []

        for deg_name, deg_data in test_data_group:
            df_emb = self.get_embeddings_csv(self.model, deg_data['filepath_deg'], root=self.config['test_mono_wav']) 

            test_embeddings = df_emb.set_index('filepath_deg')
            test_names = df_emb.merge(deg_data, on='filepath_deg')[['filepath_deg', 'Condition']]         

            euclid_dist = cdist(test_embeddings, ref_embeddings)
            avg_dist_nmr = np.mean(euclid_dist, axis=1)
            names = test_embeddings.index

            # Add distances 
            df_dist = pd.DataFrame({'filepath_deg': names, 'Distance': avg_dist_nmr})
            df_dist = df_dist.merge(test_names, on='filepath_deg')
            df_dist = df_dist.groupby('Condition').mean().reset_index()
            df_dist.sort_values(by='Distance', inplace=True)
            
            # Store for scatterplot
            degr_names.append([deg_name]*len(df_dist))
            degr_data.append(df_dist)

            # Performance Evaluation    
            SRCC, _ = spearmanr(df_dist['Distance'], df_dist['Condition'])
            srcc_scores.append([SRCC]*len(df_dist))
            print(f'Degradation: {deg_name}')
            print(f'SRCC: {np.round(SRCC, 2)}')

    def eval_full_reference(self, model_path):
        self.model.load_state_dict(torch.load(model_path))
        self.model.eval()

        # Get dataframe test set
        test_data = pd.read_csv(self.config['test_db_file_fr'])

        # Loop over each database
        db_groups = test_data.groupby('db')
        
        for db_name, db in db_groups:
            print(db_name)
            df_emb_ref = self.get_embeddings_csv(self.model, db['filepath_ref'], root=self.config['test_root_wav']).set_index('filepath_ref')
            df_emb_test = self.get_embeddings_csv(self.model, db['filepath_deg'], root=self.config['test_root_wav']).set_index('filepath_deg')
            test_names = df_emb_test.merge(db, on='filepath_deg')[['filepath_deg', 'condition', 'mos']]   
            
            # Eucl distance only with matching reference (take diagonal)
            euclid_dist = cdist(df_emb_test, df_emb_ref)
            fr_distance = np.diag(euclid_dist)
            names = df_emb_test.index

            # # Add distances 
            df_dist = pd.DataFrame({'filepath_deg': names, 'Distance': fr_distance})
            df_dist = df_dist.merge(test_names, on='filepath_deg')
            df_dist = df_dist.groupby('condition').mean()

            # Compute third order poly mapping (paper results are reported without mapping)
            #df_grouped = df_dist.groupby('condition')['mos', 'Distance'].mean()
            popt3, _ = curve_fit(self.order_three, df_dist['Distance'].values, df_dist['mos'].values)
            a3, b3, c3, d3 = popt3
            df_dist['Distance_map'] = df_dist['Distance'].apply(lambda x: self.order_three(x,a3,b3,c3,d3))

            # Scatter plot MOS - Embedding Distances
            sns.scatterplot(data=df_dist, x='mos', y='Distance_map')
            plt.xlabel('Actual MOS')
            plt.ylabel(f'Dist w.r.t Reference')
            plt.xlim([1, 5])
            plt.ylim([1, 5])
            plt.tight_layout()
            out_dir = '/'.join(self.config['nomad_model_path'].split('/')[:-1])
            plt.savefig(os.path.join(out_dir, f'fr_{db_name}_embeddings.png'))
            plt.close()
            
            # Performance Evaluation Spearman
            SRCC, _ = spearmanr(df_dist['Distance'], df_dist['mos'])
            print(f'SRCC: {np.round(SRCC, 2)}')
            SRCC, _ = spearmanr(df_dist['Distance_map'], df_dist['mos'])
            print(f'SRCC 3rd map: {np.round(SRCC, 2)}')

            # Performance Evaluation Pearson
            PCC, _ = pearsonr(df_dist['Distance'], df_dist['mos'])
            print(f'PCC: {np.round(PCC, 2)}')
            PCC, _ = pearsonr(df_dist['Distance_map'], df_dist['mos'])
            print(f'PCC 3rd map: {np.round(PCC, 2)}')
    
    def get_nmr_embeddings(self):
        # Dataframe
        ref_files = pd.DataFrame(os.listdir(self.config['non_match_dir']))
        ref_files.columns = ['reference']
        ref_files['reference'] = [os.path.join(self.config['non_match_dir'], x) for x in ref_files['reference']]
        
        # Calculate embeddings
        ref_embeddings = self.get_embeddings_csv(self.model, ref_files)
        return ref_embeddings
    
    # Use this function to check if cdist gives the same result
    def euclidean_dist(self, emb_a, emb_b):
        dist = np.sqrt(np.dot(emb_a - emb_b, (emb_a - emb_b).T))
        return dist


import pandas as pd
import random
from sklearn.model_selection import GroupShuffleSplit
import numpy as np
from itertools import combinations
from pathlib import Path
import os

# Set seed
# Seed is used for the split but not for sampling positive and negative whereas the global seed has been used
# You find the csv in the repo if you want to reuse the same train/valid sets. Otherwise specify the random_state in the sample function.
SEED = 10
random.seed(SEED)

# Import dataframe
path = '/media/alergn/hdd/github/speech-degradator/degraded_data_visqol_scores.csv'
df = pd.read_csv(path)
df.drop(index=df.index[:2], inplace=True)
df.dropna(axis=1, inplace=True)
df['nsim'] = df.iloc[:,3:35].mean(axis=1)
df = df[['reference', 'degraded', 'nsim']]
df['degraded'] = ['/'.join(x.split('/')[-2:]) for x in df['degraded']]

# Split train and val based on the clean file (never overlap the same clean file between the partitions)
splitter = GroupShuffleSplit(test_size=.20, n_splits=1, random_state = SEED)
split = splitter.split(df, groups=df['reference'])
train_inds, test_inds = next(split)
train_df = df.iloc[train_inds]
val_df = df.iloc[test_inds]

def create_triplets(df, N=1, hard_sampling=True):
    margin = 0.05
    df.drop_duplicates(inplace=True)
    anchor_list, positive_list, negative_list = [], [], []
    positive_nsim, negative_nsim = [], []
    positive_dist, negative_dist = [], []
    
    # Create Triplets (Anchor, Positive, Negative)
    for row in df['reference'].unique():
        df_g = df[df['reference'] == row]
        parts = Path(df_g.iloc[0]['reference']).parts
        filepath_clean = os.path.join('CLEAN', '/'.join(str(Path(*parts[parts.index('train-clean-100-wav'):])).split('/')[1:]))
        df_clean = pd.DataFrame({'degraded': filepath_clean, 'nsim': [1.0000]})
        df_g = pd.concat([df_g, df_clean])
        df_g.drop('reference', axis=1, inplace=True)
        
        for _ in range(N):
            # Sample anchor
            anchor = df_g.sample(1)
            anchor['nsim_dist'] = 0.0
            positive = df_g.iloc[(df_g['nsim']-anchor['nsim'].values).abs().argsort()[:2]].iloc[1].to_frame().T
            positive['nsim_dist'] = abs(anchor['nsim'].values - positive['nsim'].values)

            # Find negative
            df_g['nsim_dist'] = abs(anchor['nsim'].values - df_g['nsim'])
            df_g.drop(positive.index, axis=0, inplace=True)
            df_g.drop(anchor.index, axis=0, inplace=True)

            # Hard sampling takes the value closest to the positive
            if not hard_sampling:
                negative = df_g[df_g['nsim_dist'] > positive['nsim_dist'].values[0] + margin].sample(1)
            else:
                negative = df_g[df_g['nsim_dist'] == df_g['nsim_dist'].min()]
            
            assert positive['nsim_dist'].values[0] < negative['nsim_dist'].values[0]
            
            # Save data
            anchor_list.append(anchor['degraded'].values[0])
            positive_list.append(positive['degraded'].values[0])
            negative_list.append(negative['degraded'].values[0])
            
            #anchor_nsim.append(anchor['nsim'].values[0])
            positive_nsim.append(positive['nsim'].values[0])
            negative_nsim.append(negative['nsim'].values[0])

            #anchor_dist.append(anchor['nsim_dist'].values[0])
            positive_dist.append(positive['nsim_dist'].values[0])
            negative_dist.append(negative['nsim_dist'].values[0])
        
    df_out = pd.DataFrame({'Anchor': anchor_list, 'Positive': positive_list, 'Negative': negative_list, 'anc_pos_dist': positive_dist, 'anc_neg_dist': negative_dist})
    return df_out

# Create datasets (Easy)
train_triplets_df = create_triplets(train_df, N=3, hard_sampling=False)
train_triplets_df.dropna(axis=0, inplace=True)
valid_triplets_df = create_triplets(val_df, N=3, hard_sampling=False)
valid_triplets_df.dropna(axis=0, inplace=True)

# Create datasets (Hard)
train_triplets_df = create_triplets(train_df, N=3, hard_sampling=True)
train_triplets_df.dropna(axis=0, inplace=True)
valid_triplets_df = create_triplets(val_df, N=3, hard_sampling=True)
valid_triplets_df.dropna(axis=0, inplace=True)


import yaml
import os
import subprocess
from degradations import mp3, opus, noise, clip_signal, vorbis, reverb
import degradations
import pandas as pd
import random
random.seed(0)

# Function that picks a random clean file
def getRandomFile(path):
    randomDir = random.choice([(x) for x in list(os.scandir(path)) if x.is_dir()]).name
    randomDir2 = random.choice([f for f in list(os.scandir(os.path.join(path, randomDir)))]).name
    listOfFiles = os.listdir(os.path.join(path, randomDir, randomDir2))
    listOfFiles = [x for x in listOfFiles if x.endswith('.flac')]
    randomFile = random.choice(listOfFiles)
    path_randomFile = os.path.join(path, randomDir, randomDir2, randomFile)
    return path_randomFile, randomFile, os.path.join(randomDir, randomDir2)

with open('src/config/config_audio_degrader.yaml') as file:
    config = yaml.load(file, Loader=yaml.FullLoader)

sr = config['sr']
in_dir = os.path.join(config['root'], config['in_dir_test'])
in_dir_wav = os.path.join(config['root'], config['in_dir_test_wav'])
out_dir = os.path.join(config['root'], config['out_dir_test'])

n_cond = 60
num_files = 4

degradations = ['MP3', 'OPUS', 'CLIP', 'NOISE', 'VORBIS', 'REVERB']

filepath_deg, filename_deg, filepath_ref, filename_ref = [], [], [], []
abs_filepath_deg, abs_filepath_ref = [], []
id_clean_file = 0

# This function is needed to have a unique clean reference for each condition
def pick_clean_file():
    # Pick random clean file
    in_filepath, filename, root = getRandomFile(in_dir)
    filename_ref = os.path.splitext(filename)[0] + '.wav'

    # Store filepath ref (relative path)
    sub_dir = root.split('/')[-2:]
    sub_dir = '/'.join(sub_dir)

    # Convert to wav for visqol
    if not os.path.isdir(os.path.join(in_dir_wav, sub_dir)):
        os.makedirs(os.path.join(in_dir_wav, sub_dir))
    
    in_filepath_wav = os.path.join(in_dir_wav, sub_dir, filename_ref)
    wav_convert = f'ffmpeg -y -i {in_filepath} -ar {sr} {in_filepath_wav}'.split(' ')
    subprocess.call(wav_convert)

    filepath_ref = os.path.join(sub_dir, filename)

    return filename_ref, filepath_ref, filename, in_filepath_wav

for nfil in range(0, num_files):

    for d in degradations:
        # *** MP3 ***
        # Create out subdir
        if d == 'MP3':
            subdir_name = 'MP3'
            out_dir_mp3 = os.path.join(out_dir, subdir_name)
            if not os.path.isdir(out_dir_mp3):
                os.makedirs(out_dir_mp3)
            
            # Call mp3 codec for all conditions
            for cond in config['mp3_test']:
                filename_ref, filepath_ref, filename, in_filepath_wav = pick_clean_file()
                # Build filename deg file
                filename_deg_file = filename.split('.')[0] + f'_{subdir_name}_{cond}.wav'
                filename_deg.append(filename_deg_file)
                
                # Build filepath deg file
                out_filepath = os.path.join(out_dir_mp3, filename_deg_file)
                filepath_deg.append(os.path.join(subdir_name, filename_deg_file))
                abs_filepath_deg.append(out_filepath)

                mp3(in_filepath_wav, out_filepath, bitrate=cond, sr=config['sr'])
                cmd_normalization = f'ffmpeg-normalize {out_filepath} -o {out_filepath} -f -ar {16000}'.split(' ')
                subprocess.call(cmd_normalization)

        elif d == 'OPUS':
            # *** OPUS ***
            # Create out subdir
            subdir_name = 'OPUS'
            out_dir_opus = os.path.join(out_dir, subdir_name)
            if not os.path.isdir(out_dir_opus):
                os.makedirs(out_dir_opus)
                    
            # Call mp3 codec for all conditions
            for cond in config['opus_test']:
                filename_ref, filepath_ref, filename, in_filepath_wav = pick_clean_file()

                filename_deg_file = filename.split('.')[0] + f'_{subdir_name}_{cond}.wav'
                filename_deg.append(filename_deg_file)

                out_filepath = os.path.join(out_dir_opus, filename_deg_file)
                filepath_deg.append(os.path.join(subdir_name, filename_deg_file))
                abs_filepath_deg.append(out_filepath)

                opus(in_filepath_wav, out_filepath, bitrate=cond, sr=config['sr'])
                cmd_normalization = f'ffmpeg-normalize {out_filepath} -o {out_filepath} -f -ar {16000}'.split(' ')
                subprocess.call(cmd_normalization)

        elif d == 'NOISE':
            # *** NOISE ***
            # Create out subdir
            subdir_name = 'NOISE'
            out_dir_noise = os.path.join(out_dir, subdir_name)
            if not os.path.isdir(out_dir_noise):
                os.makedirs(out_dir_noise)
                
            # Pick one noise file
            noise_dir = config['root_noise'] + config['noise_dir_test']
            list_noise_files = [nx for nx in os.listdir(noise_dir) if nx.endswith('.wav')]
            noise_path = os.path.join(noise_dir, random.choice(list_noise_files))
            for cond in config['noise_test']:
                filename_ref, filepath_ref, filename, in_filepath_wav = pick_clean_file()

                str_cond = str(cond)
                filename_deg_file = filename.split('.')[0] + f'_{subdir_name}_{str_cond}.wav'
                filename_deg.append(filename_deg_file)

                out_filepath = os.path.join(out_dir_noise, filename_deg_file)
                filepath_deg.append(os.path.join(subdir_name, filename_deg_file))
                abs_filepath_deg.append(out_filepath)
                
                noise(in_filepath_wav, noise_path, out_filepath, snr_db=cond, sr=config['sr'])
                cmd_normalization = f'ffmpeg-normalize {out_filepath} -o {out_filepath} -f -ar {16000}'.split(' ')
                subprocess.call(cmd_normalization)

        elif d == 'CLIP':
            # *** CLIP ***
            # Create out subdir
            subdir_name = 'CLIP'
            out_dir_clip = os.path.join(out_dir, subdir_name)
            if not os.path.isdir(out_dir_clip):
                os.makedirs(out_dir_clip)
            
            # Clip
            for cond in config['clip_test']:
                filename_ref, filepath_ref, filename, in_filepath_wav = pick_clean_file()

                str_cond = str(cond)
                filename_deg_file = filename.split('.')[0] + f'_{subdir_name}_{str_cond}.wav'
                filename_deg.append(filename_deg_file)

                out_filepath = os.path.join(out_dir_clip, filename_deg_file)
                filepath_deg.append(os.path.join(subdir_name, filename_deg_file))
                abs_filepath_deg.append(out_filepath)            

                clip_signal(in_filepath_wav, out_filepath, clip_factor=cond, sr=config['sr'])
                cmd_normalization = f'ffmpeg-normalize {out_filepath} -o {out_filepath} -f -ar {16000}'.split(' ')
                subprocess.call(cmd_normalization)
        
        elif d == 'VORBIS':
            # VOrbis
            subdir_name = 'VORBIS'
            out_dir_vorbis = os.path.join(out_dir, subdir_name)
            if not os.path.isdir(out_dir_vorbis):
                os.makedirs(out_dir_vorbis)

            for cond in config['vorbis']:
                filename_ref, filepath_ref, filename, in_filepath_wav = pick_clean_file()

                str_cond = str(cond)
                filename_deg_file = filename.split('.')[0] + f'_{subdir_name}_{str_cond}.wav'
                filename_deg.append(filename_deg_file)

                out_filepath = os.path.join(out_dir_vorbis, filename_deg_file)
                filepath_deg.append(os.path.join(subdir_name, filename_deg_file))
                abs_filepath_deg.append(out_filepath)            

                vorbis(in_filepath_wav, out_filepath, quality=cond, sr=config['sr'])
                cmd_normalization = f'ffmpeg-normalize {out_filepath} -o {out_filepath} -f -ar {16000}'.split(' ')
                subprocess.call(cmd_normalization)

        elif d == 'REVERB':
            # Reverb
            subdir_name = 'REVERB'
            out_dir_reverb = os.path.join(out_dir, subdir_name)
            if not os.path.isdir(out_dir_reverb):
                os.makedirs(out_dir_reverb)

            for cond in config['reverb']:
                filename_ref, filepath_ref, filename, in_filepath_wav = pick_clean_file()

                str_cond = str(cond)
                filename_deg_file = filename.split('.')[0] + f'_{subdir_name}_{str_cond}.wav'
                filename_deg.append(filename_deg_file)

                out_filepath = os.path.join(out_dir_reverb, filename_deg_file)
                filepath_deg.append(os.path.join(subdir_name, filename_deg_file))
                abs_filepath_deg.append(out_filepath)            

                reverb(in_filepath_wav, out_filepath, p=cond, sr=config['sr'])
                cmd_normalization = f'ffmpeg-normalize {out_filepath} -o {out_filepath} -f -ar {16000}'.split(' ')
                subprocess.call(cmd_normalization)

df = pd.DataFrame({'filename_deg': filename_deg, 'filepath_deg': filepath_deg})
df['Degradation'] = [x.split('_')[-2] for x in df['filename_deg']]
df['Condition'] = [x.split('_')[-1].split('.')[0] for x in df['filename_deg']]
df['Condition'] = [int( re.sub('[^0-9]','', x)) for x in df['Condition']]

df.to_csv('./data/test_degradation_intensity_new.csv', index=False)


import fairseq
import torch
import torch.nn as nn
import pandas as pd
import os
import numpy as np
from tqdm import tqdm
from scipy.spatial.distance import cdist
from urllib.request import urlretrieve
import torchaudio
import torch.nn.functional as F
from datetime import datetime

# *** Pytorch models download options ****
if not os.path.isdir('./pt-models'):
    os.makedirs('./pt-models')

# Download wav2vec 2.0 base
url_w2v = "https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_small.pt"
w2v_path = 'pt-models/wav2vec_small.pt'
if not os.path.isfile(w2v_path):
    urlretrieve(url_w2v, w2v_path)
    print('Model wav2vec 2.0 download completed')

# Download NOMAD
url_nomad_db = 'https://www.dropbox.com/scl/fi/uws3wk327adbwqo22cr0p/nomad_best_model.pt?rlkey=cco21iba6xxi81a0dm9lpa7zj&dl=1'
nomad_path = 'pt-models/nomad_best_model.pt'
if not os.path.isfile(nomad_path):
    urlretrieve(url_nomad_db, nomad_path)
    print('Model nomad download completed')

class Nomad():
    def __init__(self, device=None):

        # *** DEVICE SETTINGS ***
        # Automatically set based on GPU detection
        if torch.cuda.is_available:
            self.DEVICE = 'cuda'
        else:
            self.DEVICE = 'cpu'
        
        # Overwrite user choice
        if device is not None:
            self.DEVICE = device
        
        print(f'NOMAD running on: {self.DEVICE}')
        
        # *** LOAD MODEL ***
        # w2v BASE parameters
        CHECKPOINT_PATH = w2v_path
        SSL_OUT_DIM = 768
        EMB_DIM = 256

        # Load w2v BASE
        w2v_model, _, _ = fairseq.checkpoint_utils.load_model_ensemble_and_task([CHECKPOINT_PATH])
        ssl_model = w2v_model[0] 
        ssl_model.remove_pretraining_modules()

        # Create NOMAD model
        model = TripletModel(ssl_model, SSL_OUT_DIM, EMB_DIM)
        MODEL_PATH = nomad_path
        model.load_state_dict(torch.load(MODEL_PATH))  
        self.model = model
        self.model.to(self.DEVICE)
        self.model.eval()

        # Create NOMAD loss model
        self.lossnet_layers = LossNetLayers(ssl_model, SSL_OUT_DIM, EMB_DIM)
        self.lossnet_layers.to(self.DEVICE)
        
        # Freeze update (no need to update gradient only propagate)
        for p in self.lossnet_layers.parameters():
            p.requires_grad=False
        
        self.nomad_loss = NomadLoss()
        self.nomad_loss.to(self.DEVICE)
        self.nomad_loss.eval()

    def predict(self,  nmr_path='data/nmr-data', test_path='data/test-data', results_path=None):
        if nmr_path is None:
            raise Exception('nmr_path not specified, you need to pass a valid value to nmr_path')
        if test_path is None:
            raise Exception('test_path not specified, you need to pass a valide value to test_path')        
        if os.path.isdir(nmr_path) == False:
            raise Exception(f'Path to the non-matching reference files {nmr_path} does not exist')
        if os.path.isdir(test_path) == False:
            raise Exception(f'Path to the non-matching reference files {test_path} does not exist')
        
        print(f'Loading non-matching references from {nmr_path}')
        nmr_embeddings = self.get_embeddings(nmr_path).set_index('filename')
        
        print(f'Loading test files from {test_path}')
        test_embeddings = self.get_embeddings(test_path).set_index('filename')

        # Compute pairwise distance matrix
        distance_matrix = cdist(test_embeddings, nmr_embeddings)

        # Compute average NOMAD score
        avg_nomad = np.mean(distance_matrix, axis=1)

        # Return both distance matrix and mean nomad score
        test_files = [x.split('/')[-1].split('.')[0] for x in test_embeddings.index]
        df_avg_nomad = pd.DataFrame({'Test File': test_files, 'NOMAD': avg_nomad}).set_index('Test File').round(3)

        df_dm = pd.DataFrame(distance_matrix).round(3)
        df_dm['Test File'] = test_files
        df_dm.set_index('Test File', inplace=True)
        df_dm.columns = [x.split('/')[-1].split('.')[0] for x in nmr_embeddings.index]
        
        # Save results
        if results_path == None:
            now = datetime.now()
            dt_string = now.strftime("%d-%m-%Y_%H-%M-%S")
            results_avg_path = os.path.join('results-csv', dt_string)
            if os.path.isdir(results_avg_path) == False:
                os.makedirs(results_avg_path)
            results_scores_path = os.path.join('results-csv', dt_string)
            if os.path.isdir(results_scores_path) == False:
                os.makedirs(results_scores_path)
            results_avg_path = os.path.join(results_avg_path, f'{dt_string}_nomad_avg.csv')
            results_scores_path = os.path.join(results_scores_path, f'{dt_string}_nomad_scores.csv')
        else:
            results_avg_path = os.path.join(results_path, 'nomad_avg.csv')
            results_scores_path = os.path.join(results_path, 'nomad_scores.csv')

        df_avg_nomad.reset_index().to_csv(results_avg_path, index=False)
        df_dm.reset_index().to_csv(results_scores_path, index=False)
        return df_avg_nomad, df_dm

    def forward(self, estimate, clean):
        estimate_embeddings = self.lossnet_layers(estimate)
        clean_embeddings = self.lossnet_layers(clean)
        loss = self.nomad_loss(clean_embeddings, estimate_embeddings)
        return loss

    def get_embeddings(self, path):
        # Dataframe
        data = pd.DataFrame(os.listdir(path))
        data.columns = ['filename']
        data['filename'] = [os.path.join(path, x) for x in data['filename']]
        
        # Calculate embeddings
        embeddings  = self.get_embeddings_csv(self.model, data)
        return embeddings

    # Function that extract NOMAD embeddings and store them in a csv
    def get_embeddings_csv(self, model, file_names, root=False):
        file_names_arr = np.array(file_names)
        embeddings = []

        model.eval()
        with torch.no_grad():
            for i, filename_anchor in enumerate(tqdm(file_names_arr)):
                if root:
                    filepath = os.path.join(root, filename_anchor)
                else:
                    filepath = filename_anchor
                
                wave = self.load_processing(filepath, trim=False)
                lengths = None
                
                wave = wave.to(self.DEVICE)
                nomad_embeddings = model(wave, lengths)
                embeddings.append(nomad_embeddings.squeeze().cpu().detach().numpy())

            embeddings = np.array(embeddings)
            embeddings = pd.DataFrame(embeddings)
            df_emb = pd.concat([file_names.reset_index(), embeddings], axis=1).drop('index', axis=1)
        
        return df_emb
    
    # Load wave file
    def load_processing(self, filepath, target_sr=16000, trim=False):
        # Load waveform
        if isinstance(filepath, np.ndarray):
            filepath = filepath[0]
        wave, sr = torchaudio.load(filepath)
        
        # Check number of channels (MONO)
        if wave.shape[0] > 1:
            wave = ((wave[0,:] + wave[1,:])/2).unsqueeze(0)
        
        # Check resampling (16 khz)
        if sr != target_sr:
            wave = torchaudio.transforms.Resample(sr,  target_sr)(wave)
            sr = target_sr
        
        # Trim audio to 10 secs
        if trim:
            if wave.shape[1] > sr*10:
                wave = wave[:, :sr*10]
        
        return wave

class TripletModel(nn.Module):
    def __init__(self, ssl_model, ssl_out_dim, emb_dim=256):
        super(TripletModel, self).__init__()
        self.ssl_model = ssl_model
        self.ssl_features = ssl_out_dim
        self.embedding_layer = nn.Sequential(
            nn.ReLU(),
            nn.Linear(self.ssl_features, emb_dim)
        )
    
    def forward(self, wav, lengths=None):
        wav = wav.squeeze(1)  ## [batches, audio_len]
        res = self.ssl_model(wav, mask=False, features_only=True)
        x = res['x']
        x_tr = torch.mean(x, 1)
        x = self.embedding_layer(x_tr)
        x = torch.nn.functional.normalize(x, dim=1)
        return x

class LossNetLayers(nn.Module):
    def __init__(self, ssl_model, ssl_out_dim, emb_dim=256):
        super(LossNetLayers, self).__init__()
        self.ssl_model = ssl_model
        self.ssl_features = ssl_out_dim
        self.embedding_layer = nn.Sequential(
            nn.ReLU(),
            nn.Linear(self.ssl_features, emb_dim)
        )

    def forward(self, wav):
        wav = wav.squeeze(1)
        res = self.ssl_model(wav, mask=False, features_only=True)

        # Get Transformer layers
        lossnet_layers = [x[0].permute(1, 0, 2) for x in res['layer_results']]

        # Get embedding layers
        x = res['x']
        x_tr = torch.mean(x, 1)
        x = self.embedding_layer(x_tr)
        x = torch.nn.functional.normalize(x, dim=1)

        # Append layers
        lossnet_layers.append(x)
        return lossnet_layers

class NomadLoss(nn.Module):
    def __init__(self):
        super(NomadLoss, self).__init__()
        # How many layers to use (12 transformer + 1 embedding)
        self.L = 13
        self.only_embedding = False
        
    def forward(self, nomad_ref, nomad_test):
        l1_dist = 0.0

        if self.only_embedding:
            ref = nomad_ref[13]
            test = nomad_test[13]
            l1_dist = F.l1_loss(test, ref)
        else:
            # Loop over each layer and calculate L1 distance
            for i in range(self.L):
                ref = nomad_ref[i]
                test = nomad_test[i]

                # Calculate L1 loss (transformer layers -> loss in each time frame)
                l1_dist += F.l1_loss(test, ref)
        return l1_dist


import click
from nomad import nomad

@click.command()
@click.option('--nmr_path', type=str, help='Path to non-matching reference files')
@click.option('--test_path', type=str, help='Path to test files')
@click.option('--results_path', type=str, default=None, help='Used to specify a path file where to save both averaged Nomad scores csv and Nomad scores csv for each non-matching reference used. Default uses a current datetime format in results-csv.')
@click.option('--device', type=str, default=None, help='Specify device, cuda or cpu. Automatically set cuda if None and GPU is detected')
def main(nmr_path, test_path, results_path, device):
    
    # Predict nomad scores
    nomad_avg, nomad_scores = nomad.predict(nmr_path, test_path, results_path)
    print('Nomad average scores, printing top 5 test files')
    print(nomad_avg.head())

if __name__ == "__main__":
    main()


import torch
import torch.nn as nn
import torch.nn.functional as F
import yaml
import numpy as np
import random
from torch.utils.data import Dataset, DataLoader
import os
import torchaudio
from tqdm import tqdm
from datetime import datetime
from pesq import pesq_batch
from nomad import nomad

with open('src/nomad_audio/se_config.yaml') as file:
    config = yaml.load(file, Loader=yaml.FullLoader)

# REPRODUCIBILITY 
SEED = 0
torch.manual_seed(SEED)
np.random.seed(SEED)
random.seed(SEED)
def seed_worker(worker_id):
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)

g = torch.Generator()
g.manual_seed(SEED)
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False

class SpeechEnhancement():
    def __init__(self):  
        # Find device
        if torch.cuda.is_available:
            self.DEVICE = 'cuda'
        else:
            self.DEVICE = 'cpu'
    
        # Create datasets
        self.train_set, self.train_loader = self.create_dl(config['noisy_train_dir'], config['clean_train_dir'], config['train_bs'], shuffle=True)
        self.valid_set, self.valid_loader = self.create_dl(config['noisy_valid_dir'], config['clean_valid_dir'], config['valid_bs'], shuffle=False)
        self.test_set, self.test_loader = self.create_dl(config['noisy_test_dir'], config['clean_test_dir'], config['test_bs'], shuffle= False)

        # Create loss
        self.mse_loss = nn.MSELoss()

        # Create model
        self.model = Model(n_layers=12)

        # Create optimizer
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=config['lr'])

    def create_dl(self, noisy_data_dir, clean_data_dir, batch_size, shuffle):
        set = AudioDataset(noisy_data_dir, clean_data_dir)
        dl = DataLoader(set, batch_size=batch_size, shuffle=shuffle, num_workers=config['num_workers'], collate_fn=set.collate_fn)
        return set, dl

    def train(self):
        self.model.train()
        total_loss = 0.0

        for id_batch, waves in enumerate(tqdm(self.train_loader)):
            estimate = self.model(waves[0]).to(self.DEVICE)
            clean = waves[1].to(self.DEVICE)

            # Calculate loss
            loss = self.mse_loss(estimate, clean) + config['nomad_weight'] * nomad.forward(estimate, clean)

            # Update weights
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

            # Update total loss
            total_loss += loss.item()

        return total_loss / len(self.train_loader)
    
    def eval(self):
        self.model.eval()
        total_loss = 0.0
        
        with torch.no_grad():
            for id_batch, waves in enumerate(tqdm(self.valid_loader)):
                estimate = self.model(waves[0]).to(self.DEVICE)
                clean = waves[1].to(self.DEVICE)

                # Calculate loss
                loss = self.mse_loss(estimate, clean) + config['nomad_weight'] * nomad.forward(estimate, clean)
                total_loss += loss.item()

        return total_loss / len(self.valid_loader)
    
    def test(self, best_model):
        pesq_scores = []
        best_model.eval()

        with torch.no_grad():
            for id_batch, waves in enumerate(tqdm(self.test_loader)):
                estimate = best_model(waves[0])
                clean = waves[1]

                pesq_scores.append(pesq_batch(fs=config['target_sr'], ref=clean.squeeze(1).detach().numpy(), deg=estimate.squeeze(1).detach().numpy(), mode='wb'))
            
            pesq_avg = np.mean([x for x in np.concatenate(pesq_scores) if isinstance(x, float)])
        return pesq_avg
    
    def training_loop(self):
        # Start Training Loop
        now = datetime.now()
        dt_string = now.strftime("%d-%m-%Y_%H-%M-%S")
        
        self.PATH_DIR = os.path.join('se_models', dt_string)
        if not os.path.isdir(self.PATH_DIR):
            os.makedirs(self.PATH_DIR)

        # Save configuration file used
        dump_path_config = os.path.join(self.PATH_DIR, 'config.yaml')
        with open(f"{dump_path_config}", 'w') as file:
            out_config_file = yaml.dump(config, file)

        best_valid_loss = np.Inf
        EPOCHS = config['num_epochs']

        counter = 0
        for i in range(EPOCHS):
            print('\n')
            train_loss = self.train()
            valid_loss = self.eval()

            if valid_loss < best_valid_loss:
                path_best_model = os.path.join(self.PATH_DIR, 'best_model.pt')
                torch.save(self.model.state_dict(), path_best_model)
                best_valid_loss = valid_loss
                print("Saved Weights Success")
                counter = 0
                best_model = self.model
            else:
                counter += 1

            print(f"COUNTER:  {counter}/{config['patience']}")
            
            if counter > config['patience']:
                print('Stop training, counter greater than patience')
                break
            
            print(f"EPOCHS: {i+1} train_loss : {train_loss}")
            print(f"EPOCHS: {i+1} valid_loss : {valid_loss}")

            # Test set PESQ every x epochs
            if (i+1) % config['test_every'] == 0:
                print('Test set evaluation')
                pesq_avg = self.test(best_model)
                print(f"EPOCHS: {i+1} PESQ : {pesq_avg}")

class AudioDataset(Dataset):
    def __init__(self, noisy_data_dir, clean_data_dir):
        self.noisy_data_dir = noisy_data_dir
        self.noisy_data = os.listdir(noisy_data_dir)
        self.clean_data_dir = clean_data_dir
        self.clean_data = os.listdir(clean_data_dir)

    def __len__(self):
        return len(self.noisy_data)

    def __getitem__(self, idx):
        noisy_audio_path = os.path.join(self.noisy_data_dir, self.noisy_data[idx])
        clean_audio_path = os.path.join(self.clean_data_dir, self.clean_data[idx])
        
        assert self.clean_data[idx] == self.noisy_data[idx]

        noisy_wave, orig_sr = torchaudio.load(noisy_audio_path)
        target_sr = config['target_sr']
        if orig_sr != target_sr:
            noisy_wave = torchaudio.transforms.Resample(orig_sr, target_sr)(noisy_wave)
        
        clean_wave, orig_sr = torchaudio.load(clean_audio_path)
        target_sr = config['target_sr']
        if orig_sr != target_sr:
            clean_wave = torchaudio.transforms.Resample(orig_sr, target_sr)(clean_wave)       

        return noisy_wave, clean_wave

    # Zero pad at batch level for waveforms (wav2vec)
    def collate_fn(self, batch):  ## zero padding
        noisy_wave, clean_wave = zip(*batch)
        noisy_wave = self.zero_pad_wav(noisy_wave)
        clean_wave = self.zero_pad_wav(clean_wave)
        return noisy_wave, clean_wave

    def zero_pad_wav(self, wavs):
        wavs = list(wavs)
        #max_len = max(wavs, key = lambda x : x.shape[1]).shape[1]
        max_len = 16384 # Same value used in the U-net repo, otherwise use the max in the batch size (see max_len line above)
        output_wavs = []
        for wav in wavs:
            if wav.shape[1] < max_len:
                amount_to_pad = max_len - wav.shape[1]
                padded_wav = torch.nn.functional.pad(wav, (0, amount_to_pad), 'constant', 0)
            else:
                padded_wav = wav[:,:max_len]
            output_wavs.append(padded_wav)

        output_wavs = torch.stack(output_wavs, dim=0)
        return output_wavs


# **** Define U-Net for speech enhancement ****
# **** CODE https://github.com/haoxiangsnr/Wave-U-Net-for-Speech-Enhancement/tree/master ****
class DownSamplingLayer(nn.Module):
    def __init__(self, channel_in, channel_out, dilation=1, kernel_size=15, stride=1, padding=7):
        super(DownSamplingLayer, self).__init__()
        self.main = nn.Sequential(
            nn.Conv1d(channel_in, channel_out, kernel_size=kernel_size,
                      stride=stride, padding=padding, dilation=dilation),
            nn.BatchNorm1d(channel_out),
            nn.LeakyReLU(negative_slope=0.1)
        )

    def forward(self, ipt):
        return self.main(ipt)

class UpSamplingLayer(nn.Module):
    def __init__(self, channel_in, channel_out, kernel_size=5, stride=1, padding=2):
        super(UpSamplingLayer, self).__init__()
        self.main = nn.Sequential(
            nn.Conv1d(channel_in, channel_out, kernel_size=kernel_size,
                      stride=stride, padding=padding),
            nn.BatchNorm1d(channel_out),
            nn.LeakyReLU(negative_slope=0.1, inplace=True),
        )

    def forward(self, ipt):
        return self.main(ipt)

class Model(nn.Module):
    def __init__(self, n_layers=12, channels_interval=24):
        super(Model, self).__init__()

        self.n_layers = n_layers
        self.channels_interval = channels_interval
        encoder_in_channels_list = [1] + [i * self.channels_interval for i in range(1, self.n_layers)]
        encoder_out_channels_list = [i * self.channels_interval for i in range(1, self.n_layers + 1)]

        #          1    => 2    => 3    => 4    => 5    => 6   => 7   => 8   => 9  => 10 => 11 =>12
        # 16384 => 8192 => 4096 => 2048 => 1024 => 512 => 256 => 128 => 64 => 32 => 16 =>  8 => 4
        self.encoder = nn.ModuleList()
        for i in range(self.n_layers):
            self.encoder.append(
                DownSamplingLayer(
                    channel_in=encoder_in_channels_list[i],
                    channel_out=encoder_out_channels_list[i]
                )
            )

        self.middle = nn.Sequential(
            nn.Conv1d(self.n_layers * self.channels_interval, self.n_layers * self.channels_interval, 15, stride=1,
                      padding=7),
            nn.BatchNorm1d(self.n_layers * self.channels_interval),
            nn.LeakyReLU(negative_slope=0.1, inplace=True)
        )

        decoder_in_channels_list = [(2 * i + 1) * self.channels_interval for i in range(1, self.n_layers)] + [
            2 * self.n_layers * self.channels_interval]
        decoder_in_channels_list = decoder_in_channels_list[::-1]
        decoder_out_channels_list = encoder_out_channels_list[::-1]
        self.decoder = nn.ModuleList()
        for i in range(self.n_layers):
            self.decoder.append(
                UpSamplingLayer(
                    channel_in=decoder_in_channels_list[i],
                    channel_out=decoder_out_channels_list[i]
                )
            )

        self.out = nn.Sequential(
            nn.Conv1d(1 + self.channels_interval, 1, kernel_size=1, stride=1),
            nn.Tanh()
        )

    def forward(self, input):
        tmp = []
        o = input

        # Down Sampling
        for i in range(self.n_layers):
            o = self.encoder[i](o)
            tmp.append(o)
            # [batch_size, T // 2, channels]
            o = o[:, :, ::2]

        o = self.middle(o)

        # Up Sampling
        for i in range(self.n_layers):
            # [batch_size, T * 2, channels]
            o = F.interpolate(o, scale_factor=2, mode="linear", align_corners=True)
            # Skip Connection
            o = torch.cat([o, tmp[self.n_layers - i - 1]], dim=1)
            o = self.decoder[i](o)

        o = torch.cat([o, input], dim=1)
        o = self.out(o)
        return o

# Testing speech enhancement
se = SpeechEnhancement()
se.training_loop()